# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Operators for manipulating tensors.
"""

from __future__ import print_function as _print_function

import sys as _sys

from tensorflow.python.ops.array_ops import gather_nd
from tensorflow.python.ops.array_ops import reshape
from tensorflow.python.ops.gen_array_ops import batch_to_space_nd
from tensorflow.python.ops.gen_array_ops import reverse_v2 as reverse
from tensorflow.python.ops.gen_array_ops import scatter_nd
from tensorflow.python.ops.gen_array_ops import space_to_batch_nd
from tensorflow.python.ops.gen_array_ops import tile
from tensorflow.python.ops.manip_ops import roll

del _print_function
