/* Copyright 2019 The TensorFlow Authors. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================*/

#ifndef TENSORFLOW_CORE_KERNELS_DATA_SERIALIZATION_UTILS_H_
#define TENSORFLOW_CORE_KERNELS_DATA_SERIALIZATION_UTILS_H_

#include "tensorflow/core/framework/dataset.h"
#include "tensorflow/core/lib/core/status.h"

namespace tensorflow {
namespace data {

// Returns a GraphDef representation of the given dataset.
Status AsGraphDef(OpKernelContext* ctx, const DatasetBase* dataset,
                  SerializationContext&& serialization_ctx,
                  GraphDef* graph_def);

// Returns a GraphDef representation of the given dataset using the minimal
// serialization parameters (i.e. ignoring external state, not serializing
// data tensors, not failing if there are datasets which do not have AsGraphDef
// implemented). Sets the `dataset_node` parameter to the dataset's
// node name in the resulting GraphDef.
Status AsGraphDefMinimal(OpKernelContext* ctx, const DatasetBase* input,
                         std::vector<std::pair<string, Tensor>>* input_list,
                         GraphDef* result, string* dataset_node);

}  // namespace data
}  // namespace tensorflow

#endif  // TENSORFLOW_CORE_KERNELS_DATA_SERIALIZATION_UTILS_H_
