# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Constants for SavedModel save and restore operations.


"""

from __future__ import print_function as _print_function

import sys as _sys

from tensorflow.python.saved_model.constants import ASSETS_DIRECTORY
from tensorflow.python.saved_model.constants import ASSETS_KEY
from tensorflow.python.saved_model.constants import DEBUG_DIRECTORY
from tensorflow.python.saved_model.constants import DEBUG_INFO_FILENAME_PB
from tensorflow.python.saved_model.constants import LEGACY_INIT_OP_KEY
from tensorflow.python.saved_model.constants import MAIN_OP_KEY
from tensorflow.python.saved_model.constants import SAVED_MODEL_FILENAME_PB
from tensorflow.python.saved_model.constants import SAVED_MODEL_FILENAME_PBTXT
from tensorflow.python.saved_model.constants import SAVED_MODEL_SCHEMA_VERSION
from tensorflow.python.saved_model.constants import VARIABLES_DIRECTORY
from tensorflow.python.saved_model.constants import VARIABLES_FILENAME

del _print_function
